/*
    
                    The Hacker's Diet

                     Database Manager               
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "astro.h"
#include "database.h"           

#define CompressedComment   32        /* Compressed comment flag in day number */

/*  DaysInMonth  --  Calculate number of days in a given month and year.  */

int DaysInMonth(int month, int year)
{
    double jd1, jd2;

    jd1 = ucttoj(year, month - 1, 1, 0, 0, 0);
    jd2 = ucttoj(month == 12 ? year + 1 : year, month == 12 ? 0 : month, 1, 0, 0, 0);

    return (int) ((jd2 + 0.1) - jd1);
}

/*  DayOfWeek  --  Determine day of week for a given date.  */

static int DayOfWeek(int month, int day, int year)
{
    return jwday(ucttoj(year, month - 1, day, 0, 0, 0));
}

/*  updateTrend  --  Calculate the day per day trend in an unpacked month  */

static void updateTrend(struct hdMonth *m)
{
    UInt i;
    trend t = m->trendCarryForward;
    signed long nt;

#ifdef NEEDED

    /*  If trend carryforward is zero, try to fill it in from the
        data of the last month.  If that fails, use the first supplied
        date in this month, but don't set the permanent carryforward,
        as we may later obtain a more accurate value.  */
        
    if (t == 0) {
        int iter = -1, bnrec = -1;
        UInt month, year, nrec = 0;
        struct hdMonth *lm;
        
        while (true) {
            hdIterateMonths(&iter, &year, &month);
            if (iter == -1) {
                break;
            }
            if ((year > m->year) || ((year == m->year) && (month >= m->month))) {
                iter = -1;
                hdIterateMonths(&iter, &year, &month);
                break;
            }
            bnrec = nrec;
            nrec++;
        }
        if (bnrec >= 0) {
            lm = hdGetMonthUnpacked(bnrec);     /* Note that this may call updateTrend recursively */
            t = m->trendCarryForward = lm->trendArray[lm->days - 1];
            free(lm);
        } else {
            for (i = 0; i < m->days; i++) {
                if (m->weightArray[i] != 0) {
                    t = WeightToTrend(m->weightArray[i]);
                    break;
                }
            }
        }
    }   
#else
    if (t == 0) {
        for (i = 0; i < m->days; i++) {
            if (m->weightArray[i] != 0) {
                t = WeightToTrend(m->weightArray[i]);
                break;
            }
        }
    }
#endif
#ifdef OLDWAY
    m->trendArray[0] = t;
    for (i = 1; i < m->days; i++) {
        if (m->weightArray[i] != 0) {
            /* Note that the fact we store weights in hundredths
               of a unit means we can compute the exponentially
               smoothed moving average entirely in integers. */
            nt = ((signed long) t) + (((signed long) WeightToTrend(m->weightArray[i])) - ((signed long) t)) / 10;
            t = (trend) nt;
        }
        m->trendArray[i] = t;
    }
#else
    for (i = 0; i < m->days; i++) {
    if (m->weightArray[i] != 0) {
        /* Note that the fact we store weights in hundredths
           of a unit means we can compute the exponentially
           smoothed moving average entirely in integers. */
        nt = ((signed long) t) + (((signed long) WeightToTrend(m->weightArray[i])) - ((signed long) t)) / 10;
        t = (trend) nt;
    }
    m->trendArray[i] = t;
    }
#endif
}

/*  unpackMonthRecord  --  Unpack a database record for a month and
                           return a dynamically allocated hdMonth
                           structure containing the data.  */
                           
struct hdMonth *unpackMonthRecord(struct hdMonthlyLog *md)
{
    struct hdMonth *m = NULL;
    UInt i, j, k;
    unsigned char c;
    char *cp;
    
    m = malloc(sizeof(struct hdMonth));
    
    if (m != NULL) {
    
        /*  Transfer scalar information */
        
        m->year = md->year;
        m->month = md->month;
        m->weightUnit = md->weightUnit;
        m->trendCarryForward = md->trendCarryForward;
        m->lastModificationTime = md->lastModificationTime;
        
        m->days = DaysInMonth(m->month + 1, m->year);       /* Days in this month */
        m->firstDay = DayOfWeek(m->month + 1, 1, m->year);  /* Weekday of first day of month */
        
        /*  Copy weight and rung arrays  */
        
        for (i = 0; i < m->days; i++) {
            m->weightArray[i] = md->weightArray[i];
            m->rungArray[i] = md->rungArray[i];
        }
/*#define STONED*/
#ifdef STONED
m->weightUnit = wuStone;
m->trendCarryForward *= 2.2;
{
    for (i = 0; i < m->days; i++) {
        m->weightArray[i] *= 2.2;
    }
}
#endif
        updateTrend(m);                 /* Calculate trendArray */
        
        /*  Unpack flag bit vector into easy-to-index array  */
        
        for (i = j = k = 0; i <= m->days; i++) {
            if (j == 0) {
                c = md->flagsArray[k++];
            }
            j = (j + 1) & 7;
            m->flagsArray[i] = (c & 0x80) ? 1 : 0;
            c <<= 1;
        }
        
        /*  Clear out comment pointer array  */
        
        memset(m->commentArray, 0, (m->days) * sizeof(unsigned char *));
        
        /*  Walk through comment text, transfer comments to our
            commentData table and set pointers to the proper days
            in commentArray.  Compressed duplicate comments are
            expanded into duplicate dynamically allocated buffers.
            We have plenty of memory for this on the desktop, and
            it makes storage management a lot simpler since we know
            that all pointers in the unpacked commentArray are
            referenced only once.  */

        i = 0;
        while (md->commentArray[i] != 0) {
            if ((md->commentArray[i] & CompressedComment) != 0) {
                m->commentArray[(md->commentArray[i] & (CompressedComment - 1)) - 1] =
                    strdup(m->commentArray[md->commentArray[i + 1] - 1]);
                i += 2;
            } else {
                cp = m->commentArray[md->commentArray[i] - 1] =
                     malloc(md->commentArray[i + 1] + 1);
                if (cp != NULL) {
                    memcpy(cp, (md->commentArray) + i + 2, md->commentArray[i + 1]);
                    cp += md->commentArray[i + 1];
                    *cp++ = 0;
                    i += md->commentArray[i + 1] + 2;
                }
            }
        }           
    }
    
    return m;
}                        

/*  packMonthRecord  --  Pack data for a month into a compact database
                         record for storage.  */
                         
struct hdMonthlyLog *packMonthRecord(struct hdMonth *m, UInt *ml)
{
    struct hdMonthlyLog *md;
    UInt d, i, j, k, commentLength;
    unsigned char f;
    char *cmt;
    /*  The compressed array for a given day is 0 if there is no
        comment for that day or the text of the comment does not
        appear in an earlier day or the day number (1 - 31) of a
        prior day whose comment is identical to this one.  */
    char compressed[32];
    
    /*  Walk through the comments and determine their packed size  */
    
    memset(compressed, 0, sizeof compressed);
    commentLength = 0;
    for (i = 0; i < m->days; i++) {
        if (m->commentArray[i] != NULL) {
            int j;
            
            for (j = 0; j < ((int) i); j++) {
                if ((m->commentArray[j] != NULL) &&
                    (strcmp(m->commentArray[i], m->commentArray[j]) == 0)) {
                    compressed[i] = (char) (j + 1);
                    break;
                }
            }
            
            if (compressed[i] != 0) {
                commentLength += 2;
            } else {
                commentLength += strlen(m->commentArray[i]) + 2;
            }
        }
    }
    
    /*  Allocate monthly log record  */
    
    md = malloc(*ml = (sizeof(struct hdMonthlyLog) + commentLength));
    
    if (md != NULL) {
        
        /*  Transfer scalar values  */
        
        md->year = m->year;
        md->month = m->month;
        md->weightUnit = m->weightUnit;
        md->trendCarryForward = m->trendCarryForward;
        md->lastModificationTime = m->lastModificationTime;
        
        /*  Copy weight and rung arrays  */
        
        for (i = 0; i < m->days; i++) {
            md->weightArray[i] = m->weightArray[i];
            md->rungArray[i] = m->rungArray[i];
        }
        
        /*  Pack flag array into a bit vector  */
        
        k = 0;
        for (i = 0; i < 4; i++) {
            f = 0;
            for (j = 0; j < 8; j++) {
                d = (k < m->days) ? m->flagsArray[k] : 0;
                f = (f << 1) | (d != 0);
                k++;    
            }
            md->flagsArray[i] = f;
        }
        
        /*  Pack comments into comment vector  */
        
        cmt = md->commentArray;
        for (i = 0; i < m->days; i++) {
            if (m->commentArray[i] != NULL) {
                if (compressed[i] != 0) {
                    *cmt++ = (char) (CompressedComment | (i + 1));
                    *cmt++ = compressed[i];
                } else {
                    UInt l = strlen(m->commentArray[i]);
                    
                    *cmt++ = (char) (i + 1);
                    *cmt++ = (char) l;
                    memcpy(cmt, m->commentArray[i], l);
                    cmt += l;
                }
            }
        }
        *cmt++ = 0;
    }
    
    return md;
}

/*  newMonth  --  Create new blank unpacked month for a
                  given year and month  */

struct hdMonth *newMonth(int year, int month)
{
    struct hdMonth *m;
    UInt i;
    
    m = malloc(sizeof(struct hdMonth));
    if (m == 0) {
        return NULL;
    }

    m->year = year;
    m->month = month;
    
    m->trendCarryForward = 0;       /* Clear trend carry forward (caller must fill in) */
    m->weightUnit = wKilogram;      /* Fill in later if you want something else */
    m->lastModificationTime = time(NULL) + timeOffset;
        
    m->days = DaysInMonth(m->month + 1, m->year);       /* Days in this month */
    m->firstDay = DayOfWeek(m->month + 1, 1, m->year);  /* Weekday of first day of month */
        
    /*  Initialise arrays for days  */
        
    for (i = 0; i < m->days; i++) {
        m->weightArray[i] = 0;
        m->rungArray[i] = RUNG_ABSENT;
        m->flagsArray[i] = 0;
        m->commentArray[i] = NULL;
    }
    
    return m;
}

/*  disposeMonth  --  Release month buffer and associated storage.  */

void disposeMonth(struct hdMonth *m)
{
    unsigned int i;

    for (i = 0; i < m->days; i++) {
        if (m->commentArray[i] != NULL) {
            free(m->commentArray[i]);
        }
    }
    free(m);
}
